/*
 * $Date$
 * 
 * Copyright (c) OSGi Alliance (2008). All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package aQute.bnd.plugin;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;

import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.*;
import org.eclipse.jdt.core.*;
import org.osgi.framework.*;
import org.osgi.util.tracker.*;

import aQute.bnd.build.*;
import aQute.bnd.classpath.*;
import aQute.bnd.service.*;
import aQute.bnd.service.action.*;

@SuppressWarnings("unchecked")
public class Central implements IResourceChangeListener {
    static IWorkspace                                iworkspace;
    final Map<IJavaProject, Project>                 javaProjectToModel = new HashMap<IJavaProject, Project>();
    final List<ModelListener>                        listeners          = new CopyOnWriteArrayList<ModelListener>();
    Workspace                                        workspace;
    final Map<RepositoryPlugin, ServiceRegistration> repoServices       = new ConcurrentHashMap<RepositoryPlugin, ServiceRegistration>();
    final ServiceTracker                             actions;
    final BundleContext                              context;
    final Workspace                                  ws;

    Central(BundleContext context) throws Exception {
        this.context = context;
        // Add a resource change listener if this is
        // the first project
        iworkspace = ResourcesPlugin.getWorkspace();
        iworkspace.addResourceChangeListener(this);

        ws = getWorkspace();
        context.registerService(Workspace.class.getName(), ws, null);
        actions = new ServiceTracker(Activator.getDefault().getBundleContext(),
                Action.class.getName(), null) {

            public Object addingService(ServiceReference ref) {

                String[] actions = (String[]) ref
                        .getProperty(Action.ACTION_MENU);
                for (String action : actions) {
                    ws.addCommand(action, new ActionWrapper(context, ref));
                }
                return ref;
            }

            public void removedService(ServiceReference ref, Object service) {
                String[] actions = (String[]) ref
                        .getProperty(Action.ACTION_MENU);
                for (String action : actions) {
                    ws.removeCommand(action);
                }
            }
        };
        actions.open();
        changed(ws);
    }

    public Project getModel(IJavaProject project) {
        try {
            Project model = javaProjectToModel.get(project);
            if (model == null) {
                File projectDir = project.getProject().getLocation()
                        .makeAbsolute().toFile();
                model = Workspace.getProject(projectDir);
                if (workspace == null) {
                    model.getWorkspace();
                }
                if (model != null) {
                    javaProjectToModel.put(project, model);
                }
            }
            return model;
        } catch (Exception e) {
            // TODO do something more useful here
            throw new RuntimeException(e);
        }
    }

    /**
     * Implementation of the resource changed interface. We are checking in the
     * POST_CHANGE phase if one of our tracked models needs to be updated.
     */
    public synchronized void resourceChanged(IResourceChangeEvent event) {
        if (event.getType() != IResourceChangeEvent.POST_CHANGE)
            return;

        IResourceDelta rootDelta = event.getDelta();
        try {
            final Set<Project> changed = new HashSet<Project>();
            rootDelta.accept(new IResourceDeltaVisitor() {
                public boolean visit(IResourceDelta delta) throws CoreException {
                    try {

                        IPath location = delta.getResource().getLocation();
                        if (location == null) {
                            System.out
                                    .println("Cannot convert resource to file: "
                                            + delta.getResource());
                        } else {
                            File file = location.toFile();
                            File parent = file.getParentFile();
                            boolean parentIsWorkspace = parent
                                    .equals(getWorkspace().getBase());

                            // file
                            // /development/osgi/svn/build/org.osgi.test.cases.distribution/bnd.bnd
                            // parent
                            // /development/osgi/svn/build/org.osgi.test.cases.distribution
                            // workspace /development/amf/workspaces/osgi
                            // false

                            if (parent != null && parentIsWorkspace) {
                                // We now are on project level, we do not go
                                // deeper
                                // because projects/workspaces should check for
                                // any
                                // changes.
                                // We are careful not to create unnecessary
                                // projects
                                // here.
                                if (workspace.isPresent(file.getName())) {
                                    Project project = workspace.getProject(file
                                            .getName());
                                    changed.add(project);
                                } else {
                                    if (file.getName().equals(Workspace.CNFDIR)) {
                                        if (workspace.refresh()) {

                                            changed.addAll(workspace
                                                    .getCurrentProjects());
                                            changed(workspace);
                                        }
                                    } else {
                                        ; // Project not created yet, so we
                                        // have
                                        // no cached results
                                    }
                                }
                                return false;
                            }
                        }
                        return true;
                    } catch (Exception e) {
                        e.printStackTrace();
                        throw new CoreException(new Status(Status.ERROR,
                                Activator.PLUGIN_ID,
                                "During checking project changes", e));
                    }
                }

            });

            for (Project p : changed) {
                if (p.refresh()) {
                    changed(p);
                }
            }
        } catch (CoreException e) {
            Activator.getDefault().error("While handling changes", e);
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /**
     * The workspace has changed. Make sure any registered services are updated.
     * 
     * @param workspace
     */
    private void changed(Workspace workspace) {
        // Get the list of repos that are listed in the workspace
        List<RepositoryPlugin> newRepos = workspace
                .getPlugins(RepositoryPlugin.class);

        // Calcalute the set of deleted by removing all current - the newRepos
        Set<RepositoryPlugin> delete = new HashSet<RepositoryPlugin>(
                repoServices.keySet());
        delete.removeAll(newRepos);

        // Remove each repo from the service registry that is no longer there
        for (RepositoryPlugin repo : delete) {
            ServiceRegistration reg = repoServices.remove(repo);
            reg.unregister();
        }

        // Calculate the new ones which is new - current
        newRepos.removeAll(repoServices.keySet());
        for (RepositoryPlugin repo : newRepos) {
            ServiceRegistration reg = Activator.getDefault().getBundleContext()
                    .registerService(RepositoryPlugin.class.getName(), repo,
                            null);
            repoServices.put(repo, reg);
        }
    }

    public Workspace getWorkspace() throws Exception {
        if (workspace != null)
            return workspace;

        IResource resource = iworkspace.getRoot().findMember("/cnf/build.bnd");
        if (resource != null) {
            IPath path = resource.getLocation();
            if (path != null) {
                File f = path.toFile();
                workspace = Workspace.getWorkspace(f.getAbsoluteFile()
                        .getParentFile().getParentFile().getAbsoluteFile());
                return workspace;
            }
        }

        workspace = Workspace.getWorkspace(iworkspace.getRoot().getLocation()
                .toFile());
        return workspace;
    }

    public void changed(Project model) {
        model.setChanged();
        for (ModelListener m : listeners)
            try {
                m.modelChanged(model);
            } catch (Exception e) {
                e.printStackTrace();
            }
    }

    public void addModelListener(ModelListener m) {
        if (!listeners.contains(m)) {
            listeners.add(m);
        }
    }

    public void removeModelListener(ModelListener m) {
        listeners.remove(m);
    }

    public IJavaProject getJavaProject(Project model) {
        for (IProject iproj : iworkspace.getRoot().getProjects()) {
            if (iproj.getName().equals(model.getName())) {
                IJavaProject ij = JavaCore.create(iproj);
                if (ij != null && ij.exists()) {
                    return ij;
                }
                // current project is not a Java project
            }
        }
        return null;
    }

    public static IPath toPath(Project project, File file) {
        String path = file.getAbsolutePath();
        String workspace = project.getWorkspace().getBase().getAbsolutePath();
        if (path.startsWith(workspace))
            path = path.substring(workspace.length());
        else
            return null;

        IPath p = new Path(path);
        return p;
    }

    public static void refresh(IPath path) {
        try {
            IResource r = ResourcesPlugin.getWorkspace().getRoot().findMember(
                    path);
            if (r != null)
                return;

            IPath p = (IPath) path.clone();
            while (p.segmentCount() > 0) {
                p = p.removeLastSegments(1);
                IResource resource = ResourcesPlugin.getWorkspace().getRoot()
                        .findMember(p);
                if (resource != null) {
                    resource.refreshLocal(2, null);
                    return;
                }
            }
        } catch (Exception e) {
            Activator.getDefault().error("While refreshing path " + path, e);
        }
        System.out.println("Unexpectedly could not find path in workspace: "
                + path);
    }

    public void refreshPlugins() throws Exception {
        List<Refreshable> rps = getWorkspace().getPlugins(Refreshable.class);
        for (Refreshable rp : rps) {
            if (rp.refresh()) {
                File dir = rp.getRoot();
                refreshFile(dir);
            }
        }
    }

    public void refreshFile(File f) throws Exception {
        String path = toLocal(f);
        IResource r = ResourcesPlugin.getWorkspace().getRoot().findMember(path);
        if (r != null) {
            r.refreshLocal(IResource.DEPTH_INFINITE, null);
        }
    }

    public void refresh(Project p) throws Exception {
        IJavaProject jp = getJavaProject(p);
        if (jp != null)
            jp.getProject().refreshLocal(IResource.DEPTH_INFINITE, null);
    }

    private String toLocal(File f) throws Exception {
        String root = getWorkspace().getBase().getAbsolutePath();
        String path = f.getAbsolutePath().substring(root.length());
        return path;
    }

    public void close() {
        actions.close();
    }
}
